#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSSettings import GXDLMSSettings
from .ValueEventArgs import ValueEventArgs
from .GXDateTime import GXDateTime
from .internal._GXCommon import _GXCommon
from .enums.Command import Command
from .GXDLMS import GXDLMS
from .GXDLMSLNParameters import GXDLMSLNParameters
from .GXByteBuffer import GXByteBuffer
from .GXDLMSSNParameters import GXDLMSSNParameters
from .VariableAccessSpecification import VariableAccessSpecification
from .enums.DataType import DataType
from .enums.Conformance import Conformance

#pylint: disable=bad-option-value,useless-object-inheritance
class GXDLMSNotify(object):
    """This class is used to send data notify and push messages to the clients."""

    #
    # Constructor.
    #
    # @param useLogicalNameReferencing
    #            Is Logical Name referencing used.
    # @param clientAddress
    #            Server address.
    # @param serverAddress
    #            Client address.
    # @param interfaceType
    #            Object type.
    #
    def __init__(self, useLogicalNameReferencing, clientAddress, serverAddress, interfaceType):
        # DLMS settings.
        self.settings = GXDLMSSettings(True)
        self.useLogicalNameReferencing = useLogicalNameReferencing
        self.settings.clientAddress = clientAddress
        self.settings.serverAddress = serverAddress
        self.settings.interfaceType = interfaceType

    def __getConformance(self):
        return self.settings.negotiatedConformance

    def __setConformance(self, value):
        self.settings.negotiatedConformance = value

    # What kind of services are used.
    conformance = property(__getConformance, __setConformance)

    #
    # @param value
    #            Cipher interface that is used to cipher PDU.
    #
    def setCipher(self, value):
        self.settings.ipher = value

    def __getObjects(self):
        return self.settings.objects

    # Get list of meter's objects.
    objects = property(__getObjects)

    #
    #
    def __getLimits(self):
        return self.settings.limits
    # Information from the connection size that server can handle.
    limits = property(__getLimits)

    def __getMaxReceivePDUSize(self):
        return self.settings.maxPduSize

    def __setMaxReceivePDUSize(self, value):
        self.settings.maxPduSize = value

    #
    # Retrieves the maximum size of received PDU.  PDU size tells maximum size
    # of PDU packet.  Value can be from 0 to 0xFFFF.  By default the value is
    # 0xFFFF.
    #
    # @see GXDLMSClient#clientAddress
    # @see GXDLMSClient#serverAddress
    # @see GXDLMSClient#useLogicalNameReferencing
    # Maximum size of received PDU.
    #
    maxReceivePDUSize = property(__getMaxReceivePDUSize, __setMaxReceivePDUSize)

    def __getUseLogicalNameReferencing(self):
        return self.settings.getUseLogicalNameReferencing()

    def __setUseLogicalNameReferencing(self, value):
        self.settings.setUseLogicalNameReferencing(value)

    #
    # Determines, whether Logical, or Short name, referencing is used.
    # Referencing depends on the device to communicate with.  Normally, a
    # device
    # supports only either Logical or Short name referencing.  The referencing
    # is defined by the device manufacturer.  If the referencing is wrong, the
    # SNMR message will fail.
    #
    # Is Logical Name referencing used.
    #
    useLogicalNameReferencing = property(__getUseLogicalNameReferencing, __setUseLogicalNameReferencing)

    def __getPriority(self):
        return self.settings.priority

    def __setPriority(self, value):
        self.settings.priority = value

    #
    # Used Priority.
    #
    priority = property(__getPriority, __setPriority)

    def __getServiceClass(self):
        return self.settings.serviceClass

    def __setServiceClass(self, value):
        self.settings.serviceClass = value

    #
    # Used service class.
    #
    serviceClass = property(__getServiceClass, __setServiceClass)

    def __getInvokeID(self):
        return self.settings.invokeId

    def __setInvokeID(self, value):
        self.settings.invokeID = value

    #
    # Invoke ID.
    #
    invokeID = property(__getInvokeID, __setInvokeID)

    #
    # Removes the HDLC frame from the packet, and returns COSEM data only.
    #
    # @param reply
    #            The received data from the device.
    # @param data
    #            Information from the received data.
    # Is frame complete.
    #
    def getData(self, reply, data):
        return GXDLMS.getData(self.settings, reply, data, None)

    #
    # Add value of COSEM object to byte buffer.  AddData method can be used
    # with
    # GetDataNotificationMessage -method.  DLMS specification do not specify
    # the
    # structure of Data-Notification body.  So each manufacture can sent
    # different data.
    #
    # @param obj
    #            COSEM object.
    # @param index
    #            Attribute index.
    # @param buff
    #            Byte buffer.
    #
    def addData(self, obj, index, buff):
        dt = None
        e = ValueEventArgs(self.settings, obj, index, 0, None)
        value = obj.getValue(self.settings, e)
        dt = obj.getDataType(index)
        if dt == DataType.NONE and value:
            dt = _GXCommon.getDLMSDataType(value)
        _GXCommon.setData(buff, dt, value)

    #
    # Generates data notification message.
    #
    # @param time
    #            Date time.  Set Date(0) if not added.
    # @param data
    #            Notification body.
    # Generated data notification message(s).
    #
    def generateDataNotificationMessages(self, time, data):
        reply = None
        if self.useLogicalNameReferencing:
            p = GXDLMSLNParameters(self.settings, 0, Command.DATA_NOTIFICATION, 0, None, data, 0xff)
            if time is None:
                p.time = None
            else:
                p.time = GXDateTime(time)
            reply = GXDLMS.getLnMessages(p)
        else:
            p2 = GXDLMSSNParameters(self.settings, Command.DATA_NOTIFICATION, 1, 0, data, None)
            reply = GXDLMS.getSnMessages(p2)
        if self.settings.negotiatedConformance & Conformance.GENERAL_BLOCK_TRANSFER == 0 and len(reply) != 1:
            raise ValueError("Data is not fit to one PDU. Use general block transfer.")
        return reply

    #
    # Generates push setup message.
    #
    # @param date
    #            Date time.  Set to null or Date(0) if not used.
    # @param push
    #            Target Push object.
    # Generated data notification message(s).
    #
    def generatePushSetupMessages(self, date, push):
        if push is None:
            raise ValueError("push")
        buff = GXByteBuffer()
        buff.setUInt8(int(DataType.STRUCTURE))
        _GXCommon.setObjectCount(push.getPushObjectList().size(), buff)
        for it in push.getPushObjectList():
            self.addData(it.getKey(), it.value.getAttributeIndex(), buff)
        return self.generateDataNotificationMessages(date, buff)

    def generateReport(self, time, list_):
        #pylint: disable=bad-option-value,redefined-variable-type
        if not list_:
            raise ValueError("list")
        if self.useLogicalNameReferencing and len(list_) != 1:
            raise ValueError("Only one object can send with Event Notification request.")
        buff = GXByteBuffer()
        reply = None
        if self.useLogicalNameReferencing:
            for k, v in list_:
                buff.setUInt16(k.objectType)
                buff.set(_GXCommon.logicalNameToBytes(k.logicalName))
                buff.setUInt8(v)
                self.addData(k, v, buff)
            p = GXDLMSLNParameters(self.settings, 0, Command.EVENT_NOTIFICATION, 0, None, buff, 0xff)
            p.time = time
            reply = GXDLMS.getLnMessages(p)
        else:
            p = GXDLMSSNParameters(self.settings, Command.INFORMATION_REPORT, len(list_), 0xFF, None, buff)
            for k, v in list_:
                buff.setUInt8(VariableAccessSpecification.VARIABLE_NAME)
                sn = k.shortName
                sn += (v - 1) * 8
                buff.setUInt16(sn)
            _GXCommon.setObjectCount(len(list_), buff)
            for k, v in list_:
                self.addData(k, v, buff)
            reply = GXDLMS.getSnMessages(p)
        return reply



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#pylint: disable=broad-except,no-name-in-module
try:
    from enum import IntEnum
    __base = IntEnum
except Exception:
    __base = object

class AutoConnectMode(__base):
    """
    Defines the mode controlling the auto dial functionality concerning the
    timing.
    """
    #pylint: disable=too-few-public-methods

    #
    # No auto dialing,
    #
    NO_AUTO_DIALLING = 0

    #
    # Auto dialing allowed any time,
    #
    AUTO_DIALLING_ALLOWED_ANYTIME = 1

    #
    # Auto dialing allowed within the validity time of the calling window.
    #
    AUTO_DIALLING_ALLOWED_CALLING_WINDOW = 2

    #
    # Regular auto dialing allowed within the validity time of the calling
    # window; alarm initiated auto dialing allowed any time,
    #
    REGULAR_AUTO_DIALLING_ALLOWED_CALLING_WINDOW = 3

    #
    # SMS sending via Public Land Mobile Network =PLMN
    #
    SMS_SENDING_PLMN = 4

    #
    # SMS sending via PSTN.
    #
    SMS_SENDING_PSTN = 5

    #
    # Email sending.
    #
    EMAIL_SENDING = 6

    #
    # The device is permanently connected to the communication network.
    #
    PERMANENTLY_CONNECT = 101
    #
    # The device is permanently connected to the communication network.  No
    # connection possible outside the calling window.
    #
    CONNECT_WITH_CALLING_WINDOW = 102
    #
    # The device is permanently connected to the communication network.
    # Connection is possible as soon as the connect method is invoked.
    #
    CONNECT_INVOKED = 103
    #
    # The device is usually disconnected.  It connects to the communication
    # network as soon as the connect method is invoked
    #
    DISCONNECT_CONNECT_INVOKED = 104

    #
    #  =200..255) manufacturer specific modes
    #
    MANUFACTURE_SPESIFIC = 200



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .enums.Security import Security
from .objects.enums.SecuritySuite import SecuritySuite
from .CountType import CountType

 # pylint: disable=too-many-instance-attributes
class AesGcmParameter:
    def __init__(self, tag=0, systemTitle=None, blockCipherKey=None, authenticationKey=None):
        """
        Constructor.

        tag : Tag.
        systemTitle : System title.
        blockCipherKey : Block cipher key.
        authenticationKey : Authentication key.
        """
        self.tag = tag
        self.security = Security.NONE
        # Invocation counter.
        self.invocationCounter = 0
        # Used security suite.
        self.securitySuite = SecuritySuite.AES_GCM_128
        # Block cipher key.
        self.blockCipherKey = blockCipherKey
        # Authentication key.
        self.authenticationKey = authenticationKey
        # System title.
        self.systemTitle = systemTitle
        # Recipient system title.
        self.recipientSystemTitle = None
        # Count type.
        self.type_ = CountType.PACKET
        # Date time.
        self.dateTime = None
        # Other information.
        self.otherInformation = None
        # Counted tag.
        self.countTag = None
        # Key parameters.
        self.keyParameters = None
        # Key ciphered data.
        self.keyCipheredData = None
        # Ciphered content.
        self.cipheredContent = None
        # Shared secret is generated when connection is made.
        self.sharedSecret = None
        self.xml = None



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .ServiceError import ServiceError
from .ConfirmedServiceError import ConfirmedServiceError
from .enums.ApplicationReference import ApplicationReference
from .enums.VdeStateError import VdeStateError
from .enums.HardwareResource import HardwareResource
from .enums.Definition import Definition
from .enums.Access import Access
from .enums.Service import Service
from .enums.Initiate import Initiate
from .enums.LoadDataSet import LoadDataSet
from .enums.Task import Task

class GXDLMSConfirmedServiceError(Exception):
    """
    DLMS specific exception class that has error description available from getDescription method.
    """

    #
    # Constructor for Confirmed ServiceError.
    #
    # @param service
    # @param type
    # @param value
    #
    def __init__(self, service=None, type_=None, value=0):
        Exception.__init__(self, "ServiceError " + self.__getConfirmedServiceError(service) + " exception. " + self.__getServiceError(type_) + " " + self.__getServiceErrorValue(type_, value))
        self.confirmedServiceError = service
        self.serviceError = type_
        self.serviceErrorValue = value

    @classmethod
    def __getConfirmedServiceError(cls, stateError):
        str_ = ""
        if stateError == ConfirmedServiceError.INITIATE_ERROR:
            str_ = "Initiate Error"
        elif stateError == ConfirmedServiceError.READ:
            str_ = "Read"
        elif stateError == ConfirmedServiceError.WRITE:
            str_ = "Write"
        return str_

    @classmethod
    def __getServiceError(cls, error):
        str_ = ""
        if error == ServiceError.APPLICATION_REFERENCE:
            str_ = "Application reference"
        elif error == ServiceError.HARDWARE_RESOURCE:
            str_ = "Hardware resource"
        elif error == ServiceError.VDE_STATE_ERROR:
            str_ = "Vde state error"
        elif error == ServiceError.SERVICE:
            str_ = "Service"
        elif error == ServiceError.DEFINITION:
            str_ = "Definition"
        elif error == ServiceError.ACCESS:
            str_ = "Access"
        elif error == ServiceError.INITIATE:
            str_ = "Initiate"
        elif error == ServiceError.LOAD_DATASET:
            str_ = "Load dataset"
        elif error == ServiceError.TASK:
            str_ = "Task"
        elif error == ServiceError.OTHER_ERROR:
            str_ = "Other Error"
        return str_

    @classmethod
    def __getServiceErrorValue(cls, error, value):
        str_ = ""
        if error == ServiceError.APPLICATION_REFERENCE:
            str_ = str(ApplicationReference(value))
        elif error == ServiceError.HARDWARE_RESOURCE:
            str_ = str(HardwareResource(value))
        elif error == ServiceError.VDE_STATE_ERROR:
            str_ = str(VdeStateError(value))
        elif error == ServiceError.SERVICE:
            str_ = str(Service(value))
        elif error == ServiceError.DEFINITION:
            str_ = str(Definition(value))
        elif error == ServiceError.ACCESS:
            str_ = str(Access(value))
        elif error == ServiceError.INITIATE:
            str_ = str(Initiate(value))
        elif error == ServiceError.LOAD_DATASET:
            str_ = str(LoadDataSet(value))
        elif error == ServiceError.TASK:
            str_ = str(Task(value))
        elif error == ServiceError.OTHER_ERROR:
            str_ = str(value)
        return str_



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..enums import ObjectType, DataType
from .enums import BaudRate, AddressState

# pylint: disable=too-many-instance-attributes
class GXDLMSMBusSlavePortSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSMBusSlavePortSetup
    """

    #
    # Constructor.
    #
    # @param ln
    # Logical Name of the object.
    # @param sn
    # Short Name of the object.
    #
    def __init__(self, ln=None, sn=0):
        super(GXDLMSMBusSlavePortSetup, self).__init__(ObjectType.MBUS_SLAVE_PORT_SETUP, ln, sn)
        # Defines the baud rate for the opening sequence.
        self.defaultBaud = BaudRate.BAUDRATE_300
        # Defines the baud rate for the opening sequence.
        self.availableBaud = BaudRate.BAUDRATE_300
        # Defines whether or not the device has been assigned an
        # address since last power up of the device.
        self.addressState = AddressState.NONE
        # Defines the baud rate for the opening sequence.
        self.busAddress = 0

    def getValues(self):
        return [self.logicalName,
                self.defaultBaud,
                self.availableBaud,
                self.addressState,
                self.busAddress]

    #
    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = list()
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  DefaultBaud
        if all_ or not self.isRead(2):
            attributes.append(2)
        #  AvailableBaud
        if all_ or not self.isRead(3):
            attributes.append(3)
        #  AddressState
        if all_ or not self.isRead(4):
            attributes.append(4)
        #  BusAddress
        if all_ or not self.isRead(5):
            attributes.append(5)
        return attributes

    def getAttributeCount(self):
        return 5

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 0

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.ENUM
        elif index == 3:
            ret = DataType.ENUM
        elif index == 4:
            ret = DataType.ENUM
        elif index == 5:
            ret = DataType.UINT16
        else:
            raise ValueError("getDataType failed. Invalid attribute index.")
        return ret
    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            return self.defaultBaud
        if e.index == 3:
            return self.availableBaud
        if e.index == 4:
            return self.addressState
        if e.index == 5:
            return self.busAddress
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        #pylint: disable=bad-option-value,redefined-variable-type
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            if e.value is None:
                self.defaultBaud = BaudRate.BAUDRATE_300
            else:
                self.defaultBaud = BaudRate(e.value)
        elif e.index == 3:
            if e.value is None:
                self.availableBaud = BaudRate.BAUDRATE_300
            else:
                self.availableBaud = BaudRate(e.value)
        elif e.index == 4:
            if e.value is None:
                self.addressState = AddressState.NONE
            else:
                self.addressState = AddressState(e.value)
        elif e.index == 5:
            if e.value is None:
                self.busAddress = 0
            else:
                self.busAddress = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.defaultBaud = BaudRate(reader.readElementContentAsInt("DefaultBaud"))
        self.availableBaud = BaudRate(reader.readElementContentAsInt("AvailableBaud"))
        self.addressState = AddressState(reader.readElementContentAsInt("AddressState"))
        self.busAddress = reader.readElementContentAsInt("BusAddress")

    def save(self, writer):
        writer.writeElementString("DefaultBaud", self.defaultBaud)
        writer.writeElementString("AvailableBaud", self.availableBaud)
        writer.writeElementString("AddressState", self.addressState)
        writer.writeElementString("BusAddress", self.busAddress)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from ..enums import Authentication

class GXAuthentication:
    """
    Authentication class is used to give authentication information to the server.
    """

    def __init__(self, forType=Authentication.NONE, pw=None, forClientAddress=0):
        """
        Constructor.
        forType: Authentication type
        pw: Used password.
        forClientAddress: Client Id.
        """

        self.type_ = forType
        if pw:
            self.password = pw[0:]
        self.clientAddress = forClientAddress

    def __str__(self):
        return str(self.type_)



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from __future__ import print_function
from .ActionRequestType import ActionRequestType
from .internal._GXCommon import _GXCommon
from .enums import Command, ErrorCode, ObjectType
from .ServiceError import ServiceError
from .enums.Service import Service
from .ConfirmedServiceError import ConfirmedServiceError
from .TranslatorTags import TranslatorTags
from .GXByteBuffer import GXByteBuffer
from .enums.AccessServiceCommandType import AccessServiceCommandType
from .internal._GXDataInfo import _GXDataInfo
from .TranslatorOutputType import TranslatorOutputType
from .SingleReadResponse import SingleReadResponse
from .GetCommandType import GetCommandType
from .GXDLMSLNParameters import GXDLMSLNParameters
from .SetRequestType import SetRequestType
from .ValueEventArgs import ValueEventArgs
from .enums.DataType import DataType
from .objects.enums.AssociationStatus import AssociationStatus
from .ConnectionState import ConnectionState
from .objects.GXDLMSAssociationLogicalName import GXDLMSAssociationLogicalName
from .GXDLMS import GXDLMS
from .GXDLMSLongTransaction import GXDLMSLongTransaction
from .enums.AccessMode import AccessMode
from .enums.MethodAccessMode import MethodAccessMode
from .objects.GXDLMSProfileGeneric import GXDLMSProfileGeneric

# pylint: disable=bad-option-value,too-many-locals,too-many-arguments,broad-except,too-many-nested-blocks,old-style-class
class GXDLMSLNCommandHandler:
    #Constructor.
    def __init__(self):
        pass

    @classmethod
    def handleGetRequest(cls, settings, server, data, replyData, xml):
        #  Return error if connection is not established.
        if xml is None and not settings.acceptConnection():
            replyData.set(server.generateConfirmedServiceError(ConfirmedServiceError.INITIATE_ERROR, ServiceError.SERVICE, Service.UNSUPPORTED))
            return
        #  Get type.
        type_ = int(data.getUInt8())
        #  Get invoke ID and priority.
        invokeID = data.getUInt8()
        settings.updateInvokeId(invokeID)
        if xml:
            xml.appendStartTag(Command.GET_REQUEST)
            xml.appendStartTag(Command.GET_REQUEST, type_)
            xml.appendLine(TranslatorTags.INVOKE_ID, "Value", xml.integerToHex(invokeID, 2))
        #  GetRequest normal
        if type_ == GetCommandType.NORMAL:
            cls.getRequestNormal(settings, invokeID, server, data, replyData, xml)
        elif type_ == GetCommandType.NEXT_DATA_BLOCK:
            #  Get request for next data block
            cls.getRequestNextDataBlock(settings, invokeID, server, data, replyData, xml, False)
        elif type_ == GetCommandType.WITH_LIST:
            #  Get request with a list.
            cls.getRequestWithList(settings, invokeID, server, data, replyData, xml)
        else:
            bb = GXByteBuffer()
            settings.resetBlockIndex()
            #  Access Error : Device reports a hardware fault.
            bb.setUInt8(ErrorCode.HARDWARE_FAULT)
            GXDLMS.getLNPdu(GXDLMSLNParameters(settings, invokeID, Command.GET_RESPONSE, type_, None, bb, ErrorCode.OK), replyData)
        if xml:
            xml.appendEndTag(Command.GET_REQUEST, type_)
            xml.appendEndTag(Command.GET_REQUEST)

    #
    # Handle set request.
    #
    # Reply to the client.
    #
    @classmethod
    def handleSetRequest(cls, settings, server, data, replyData, xml):
        #  Return error if connection is not established.
        if xml is None and not settings.acceptConnection():
            replyData.set(server.generateConfirmedServiceError(ConfirmedServiceError.INITIATE_ERROR, ServiceError.SERVICE, Service.UNSUPPORTED))
            return
        #  Get type.
        type_ = int(data.getUInt8())
        #  Get invoke ID and priority.
        invoke = data.getUInt8()
        settings.updateInvokeId(invoke)
        #  SetRequest normal or Set Request With First Data Block
        p = GXDLMSLNParameters(settings, invoke, Command.SET_RESPONSE, type_, None, None, 0)
        if xml:
            xml.appendStartTag(Command.SET_REQUEST)
            xml.appendStartTag(Command.SET_REQUEST, type_)
            #  InvokeIdAndPriority
            xml.appendLine(TranslatorTags.INVOKE_ID, "Value", xml.integerToHex(invoke, 2))
        if type_ in (SetRequestType.NORMAL, SetRequestType.FIRST_DATA_BLOCK):
            cls.handleSetRequestNormal(settings, server, data, type_, p, xml)
        elif type_ == SetRequestType.WITH_DATA_BLOCK:
            cls.hanleSetRequestWithDataBlock(settings, server, data, p, xml)
        elif type_ == SetRequestType.WITH_LIST:
            cls.hanleSetRequestWithList(settings, invoke, server, data, xml)
        else:
            settings.resetBlockIndex()
            p.status = ErrorCode.HARDWARE_FAULT
        if xml:
            xml.appendEndTag(Command.SET_REQUEST, type_)
            xml.appendEndTag(Command.SET_REQUEST)
            return
        GXDLMS.getLNPdu(p, replyData)

    @classmethod
    def appendAttributeDescriptor(cls, xml, ci, ln, attributeIndex):
        xml.appendStartTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR)
        if xml.comments:
            ot = ObjectType(ci)
            if ot:
                xml.appendComment(str(ot))
        xml.appendLine(TranslatorTags.CLASS_ID, "Value", xml.integerToHex(ci, 4))
        xml.appendComment(_GXCommon.toLogicalName(ln))
        xml.appendLine(TranslatorTags.INSTANCE_ID, "Value", GXByteBuffer.hex(ln, False))
        xml.appendLine(TranslatorTags.ATTRIBUTE_ID, "Value", xml.integerToHex(attributeIndex, 2))
        xml.appendEndTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR)

    @classmethod
    def appendMethodDescriptor(cls, xml, ci, ln, attributeIndex):
        xml.appendStartTag(TranslatorTags.METHOD_DESCRIPTOR)
        if xml.comments:
            ot = ObjectType(ci)
            if ot:
                xml.appendComment(str(ot))
        xml.appendLine(TranslatorTags.CLASS_ID, "Value", xml.integerToHex(ci, 4))
        xml.appendComment(_GXCommon.toLogicalName(ln))
        xml.appendLine(TranslatorTags.INSTANCE_ID, "Value", GXByteBuffer.hex(ln, False))
        xml.appendLine(TranslatorTags.METHOD_ID, "Value", xml.integerToHex(attributeIndex, 2))
        xml.appendEndTag(TranslatorTags.METHOD_DESCRIPTOR)

    #
    # Handle get request normal command.
    #
    # @param data
    #            Received data.
    #
    @classmethod
    def getRequestNormal(cls, settings, invokeID, server, data, replyData, xml):
        bb = GXByteBuffer()
        e = None
        status = ErrorCode.OK
        settings.setCount(0)
        settings.setIndex(0)
        settings.resetBlockIndex()
        #  CI
        ci = data.getUInt16()
        ln = bytearray(6)
        data.get(ln)
        #  Attribute Id
        attributeIndex = data.getUInt8()
        #  AccessSelection
        selection = data.getUInt8()
        selector = 0
        parameters = None
        info = _GXDataInfo()
        if selection != 0:
            selector = data.getUInt8()
        if xml:
            cls.appendAttributeDescriptor(xml, ci, ln, attributeIndex)
            if selection != 0:
                info.xml = (xml)
                xml.appendStartTag(TranslatorTags.ACCESS_SELECTION)
                xml.appendLine(TranslatorTags.ACCESS_SELECTOR, "Value", xml.integerToHex(selector, 2))
                xml.appendStartTag(TranslatorTags.ACCESS_PARAMETERS)
                _GXCommon.getData(data, info)
                xml.appendEndTag(TranslatorTags.ACCESS_PARAMETERS)
                xml.appendEndTag(TranslatorTags.ACCESS_SELECTION)
            return
        if selection != 0:
            parameters = _GXCommon.getData(data, info)
        ot = ObjectType(ci)
        obj = settings.objects.findByLN(ot, _GXCommon.toLogicalName(ln))
        if obj is None:
            obj = server.onFindObject(ot, 0, _GXCommon.toLogicalName(ln))
        e = ValueEventArgs(server, obj, attributeIndex, selector, parameters)
        if obj is None:
            #  "Access Error : Device reports a undefined object."
            status = ErrorCode.UNDEFINED_OBJECT
        else:
            e.invokeId = (invokeID)
            if server.onGetAttributeAccess(e) == AccessMode.NO_ACCESS:
                #  Read Write denied.
                status = ErrorCode.READ_WRITE_DENIED
            else:
                if isinstance(obj, (GXDLMSProfileGeneric,)) and attributeIndex == 2:
                    dt = None
                    rowsize = 0
                    pg = e.target
                    #  Count how many rows we can fit to one PDU.
                    for it in pg.getCaptureObjects():
                        dt = it.getKey().getDataType(it.value.getAttributeIndex())
                        if dt == DataType.OCTET_STRING:
                            dt = it.getKey().getUIDataType(it.value.getAttributeIndex())
                            if dt == DataType.DATETIME:
                                rowsize += _GXCommon.getDataTypeSize(DataType.DATETIME)
                            elif dt == DataType.DATE:
                                rowsize += _GXCommon.getDataTypeSize(DataType.DATE)
                            elif dt == DataType.TIME:
                                rowsize += _GXCommon.getDataTypeSize(DataType.TIME)
                        elif dt == DataType.NONE:
                            rowsize += 2
                        else:
                            rowsize += _GXCommon.getDataTypeSize(dt)
                    if rowsize != 0:
                        e.rowToPdu = int(settings.maxPduSize / rowsize)
                server.notifyRead()
                value = None
                if e.handled:
                    value = e.value
                else:
                    settings.setCount(e.rowEndIndex - e.rowBeginIndex)
                    value = obj.getValue(settings, e)
                server.onPostRead()
                if e.byteArray:
                    bb.set(int(value))
                else:
                    GXDLMS.appendData(obj, attributeIndex, bb, value)
                status = e.error
        GXDLMS.getLNPdu(GXDLMSLNParameters(settings, e.invokeId, Command.GET_RESPONSE, 1, None, bb, status), replyData)
        if settings.count != settings.index or len(bb) != bb.position:
            server.setTransaction(GXDLMSLongTransaction(e, Command.GET_REQUEST, bb))

    #
    # Handle get request next data block command.
    #
    # @param data
    #            Received data.
    #
    @classmethod
    def getRequestNextDataBlock(cls, settings, invokeID, server, data, replyData, xml, streaming):
        bb = GXByteBuffer()
        if not streaming:
            index = int(data.getUInt32())
            #  Get block index.
            if xml:
                xml.appendLine(TranslatorTags.BLOCK_NUMBER, None, xml.integerToHex(index, 8))
                return
            if index != settings.blockIndex:
                GXDLMS.getLNPdu(GXDLMSLNParameters(settings, invokeID, Command.GET_RESPONSE, 2, None, bb, ErrorCode.DATA_BLOCK_NUMBER_INVALID), replyData)
                return
        settings.increaseBlockIndex()
        p = GXDLMSLNParameters(settings, invokeID, Command.GENERAL_BLOCK_TRANSFER if streaming else Command.GET_RESPONSE, 2, None, bb, ErrorCode.OK)
        p.streaming = streaming
        p.windowSize = settings.getWindowSize()
        #  If transaction is not in progress.
        if server.getTransaction() is None:
            p.status = int(ErrorCode.NO_LONG_GET_OR_READ_IN_PROGRESS)
        else:
            bb.set(server.getTransaction().data)
            moreData = settings.index != settings.getCount()
            if moreData:
                #  If there is multiple blocks on the buffer.
                #  This might happen when Max PDU size is very small.
                if len(bb) < settings.getMaxPduSize():
                    value = None
                    for arg in server.transaction.targets:
                        arg.invokeId = (p.invokeId)
                        server.onPreRead([arg])
                        if arg.handled:
                            value = arg.value
                        else:
                            value = arg.target.getValue(settings, arg)
                        p.invokeId = (arg.invokeId)
                        #  Add data.
                        if arg.byteArray:
                            bb.set(int(value))
                        else:
                            GXDLMS.appendData(arg.target, arg.index, bb, value)
                    moreData = settings.index != settings.getCount()
            p.multipleBlocks = True
            GXDLMS.getLNPdu(p, replyData)
            if moreData or len(bb) - bb.position != 0:
                server.getTransaction().setData(bb)
            else:
                server.setTransaction(None)
                settings.resetBlockIndex()

    #
    # Handle get request with list command.
    #
    # @param data
    #            Received data.
    #
    @classmethod
    def getRequestWithList(cls, settings, invokeID, server, data, replyData, xml):
        bb = GXByteBuffer()
        pos = int()
        cnt = _GXCommon.getObjectCount(data)
        _GXCommon.setObjectCount(cnt, bb)
        list_ = list()
        if xml:
            xml.appendStartTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR_LIST, "Qty", xml.integerToHex(cnt, 2))
        while pos != cnt:
            ci = ObjectType(data.getUInt16())
            ln = bytearray(6)
            data.get(ln)
            attributeIndex = data.getUInt8()
            #  AccessSelection
            selection = data.getUInt8()
            selector = 0
            parameters = None
            if selection != 0:
                selector = data.getUInt8()
                i = _GXDataInfo()
                parameters = _GXCommon.getData(data, i)
            if xml:
                xml.appendStartTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR_WITH_SELECTION)
                xml.appendStartTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR)
                if xml.comments:
                    xml.appendComment(ci.__str__())
                xml.appendLine(TranslatorTags.CLASS_ID, "Value", xml.integerToHex(ci.value, 4))
                xml.appendComment(_GXCommon.toLogicalName(ln))
                xml.appendLine(TranslatorTags.INSTANCE_ID, "Value", GXByteBuffer.hex(ln, False))
                xml.appendLine(TranslatorTags.ATTRIBUTE_ID, "Value", xml.integerToHex(attributeIndex, 2))
                xml.appendEndTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR)
                xml.appendEndTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR_WITH_SELECTION)
            else:
                obj = settings.objects.findByLN(ci, _GXCommon.toLogicalName(ln))
                if obj is None:
                    obj = server.onFindObject(ci, 0, _GXCommon.toLogicalName(ln))
                arg = ValueEventArgs(server, obj, attributeIndex, selector, parameters)
                arg.invokeId = (invokeID)
                if obj is None:
                    arg.error = ErrorCode.UNDEFINED_OBJECT
                    list_.append(arg)
                else:
                    if server.onGetAttributeAccess(arg) == AccessMode.NO_ACCESS:
                        #  Read Write denied.
                        arg.error = ErrorCode.READ_WRITE_DENIED
                        list_.append(arg)
                    else:
                        list_.append(arg)
            pos += 1
        if xml:
            xml.appendEndTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR_LIST)
            return
        server.onPreRead(list_)
        value = None
        pos = 0
        p = GXDLMSLNParameters(settings, invokeID, Command.GET_RESPONSE, 3, None, bb, 0xFF)
        for it in list_:
            try:
                if it.handled:
                    value = it.value
                else:
                    value = it.target.getValue(settings, it)
                bb.setUInt8(it.error)
                if it.byteArray:
                    bb.set(value)
                else:
                    GXDLMS.appendData(it.target, it.index, bb, value)
                p.invokeId = it.invokeId
            except Exception:
                bb.setUInt8(ErrorCode.HARDWARE_FAULT)
            if settings.index != settings.count:
                server.setTransaction(GXDLMSLongTransaction(list_, Command.GET_REQUEST, None))
            pos += 1
        server.onPostRead(list_)
        GXDLMS.getLNPdu(p, replyData)

    @classmethod
    def handleSetRequestNormal(cls, settings, server, data, type_, p, xml):
        value = None
        reply = _GXDataInfo()
        #  CI
        ci = data.getInt16()
        ot = ObjectType(ci & 0xFFFF)
        ln = bytearray(6)
        data.get(ln)
        #  Attribute index.
        index = data.getUInt8()
        #  Get Access Selection.
        data.getUInt8()
        if type_ == 2:
            lastBlock = data.getUInt8()
            p.multipleBlocks = lastBlock == 0
            blockNumber = data.getUInt32()
            if blockNumber != settings.blockIndex:
                p.status = ErrorCode.DATA_BLOCK_NUMBER_INVALID
                return
            settings.increaseBlockIndex()
            size = _GXCommon.getObjectCount(data)
            realSize = len(data) - data.position
            if size != realSize:
                p.status = ErrorCode.DATA_BLOCK_UNAVAILABLE
                return
            if xml:
                cls.appendAttributeDescriptor(xml, ci, ln, index)
                xml.appendStartTag(TranslatorTags.DATA_BLOCK)
                xml.appendLine(TranslatorTags.LAST_BLOCK, "Value", xml.integerToHex(lastBlock, 2))
                xml.appendLine(TranslatorTags.BLOCK_NUMBER, "Value", xml.integerToHex(blockNumber, 8))
                xml.appendLine(TranslatorTags.RAW_DATA, "Value", data.remainingHexString(False))
                xml.appendEndTag(TranslatorTags.DATA_BLOCK)
            return
        if xml:
            cls.appendAttributeDescriptor(xml, ci, ln, index)
            xml.appendStartTag(TranslatorTags.VALUE)
            di = _GXDataInfo()
            di.xml = (xml)
            value = _GXCommon.getData(data, di)
            if not di.complete:
                value = GXByteBuffer.hex(data.data, False, data.position, len(data) - data.position)
            elif isinstance(value, bytearray):
                value = GXByteBuffer.hex(value, False)
            xml.appendEndTag(TranslatorTags.VALUE)
            return
        if not p.isMultipleBlocks():
            settings.resetBlockIndex()
            value = _GXCommon.getData(data, reply)
        obj = settings.objects.findByLN(ot, _GXCommon.toLogicalName(ln))
        if obj is None:
            obj = server.onFindObject(ot, 0, _GXCommon.toLogicalName(ln))
        #  If target is unknown.
        if obj is None:
            #  Device reports a undefined object.
            p.setStatus(ErrorCode.UNDEFINED_OBJECT)
        else:
            e = ValueEventArgs(server, obj, index, 0, None)
            e.invokeId = (p.invokeId)
            am = server.onGetAttributeAccess(e)
            #  If write is denied.
            if am not in (AccessMode.WRITE, AccessMode.READ_WRITE):
                #  Read Write denied.
                p.setStatus(ErrorCode.READ_WRITE_DENIED)
            else:
                try:
                    if isinstance(value, bytearray):
                        dt = obj.getDataType(index)
                        if dt not in (DataType.NONE, DataType.OCTET_STRING):
                            value = _GXCommon.changeType(value, dt)
                    e.value = value
                    list_ = list(e)
                    if p.isMultipleBlocks():
                        server.setTransaction(GXDLMSLongTransaction(list_, Command.GET_REQUEST, data))
                    server.onPreWrite(list_)
                    if e.error != ErrorCode.OK:
                        p.status = e.error
                    elif not e.handled and not p.multipleBlocks:
                        obj.setValue(settings, e)
                    server.onPostWrite(list_)
                    p.invokeId = e.invokeId
                except Exception:
                    p.setStatus(ErrorCode.HARDWARE_FAULT)

    @classmethod
    def hanleSetRequestWithDataBlock(cls, settings, server, data, p, xml):
        reply = _GXDataInfo()
        lastBlock = data.getUInt8()
        p.multipleBlocks = lastBlock == 0
        blockNumber = data.getUInt32()
        if xml is None and blockNumber != settings.blockIndex:
            p.status = ErrorCode.DATA_BLOCK_NUMBER_INVALID
        else:
            settings.increaseBlockIndex()
            size = _GXCommon.getObjectCount(data)
            realSize = len(data) - data.position
            if size != realSize:
                p.status = ErrorCode.DATA_BLOCK_UNAVAILABLE
            if xml:
                xml.appendStartTag(TranslatorTags.DATA_BLOCK)
                xml.appendLine(TranslatorTags.LAST_BLOCK, "Value", xml.integerToHex(lastBlock, 2))
                xml.appendLine(TranslatorTags.BLOCK_NUMBER, "Value", xml.integerToHex(blockNumber, 8))
                xml.appendLine(TranslatorTags.RAW_DATA, "Value", data.remainingHexString(False))
                xml.appendEndTag(TranslatorTags.DATA_BLOCK)
                return
            server.getTransaction().data.set(data)
            if not p.isMultipleBlocks():
                try:
                    value = _GXCommon.getData(server.getTransaction().data, reply)
                    if isinstance(value, bytearray):
                        dt = server.transaction.targets[0].target.getDataType(server.transaction.targets[0].index)
                        if dt not in (DataType.NONE, DataType.OCTET_STRING):
                            value = _GXCommon.changeType(value, dt)
                    server.transaction.targets[0].setValue(value)
                    server.onPreWrite(server.transaction.targets)
                    if not server.transaction.targets[0].handled and not p.isMultipleBlocks():
                        server.transaction.targets[0].target.setValue(settings, server.transaction.targets[0])
                    server.onPostWrite(server.transaction.targets)
                except Exception:
                    p.setStatus(ErrorCode.HARDWARE_FAULT)
                finally:
                    server.setTransaction(None)
                settings.resetBlockIndex()
        p.multipleBlocks = True

    @classmethod
    def hanleSetRequestWithList(cls, settings, invokeID, server, data, xml):
        e = None
        cnt = _GXCommon.getObjectCount(data)
        list_ = list()
        if xml:
            xml.appendStartTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR_LIST, "Qty", xml.integerToHex(cnt, 2))
        try:
            pos = 0
            while pos != cnt:
                ci = ObjectType(data.getUInt16())
                ln = bytearray(6)
                data.get(ln)
                attributeIndex = data.getUInt8()
                selection = data.getUInt8()
                selector = 0
                parameters = None
                if selection != 0:
                    selector = data.getUInt8()
                    info = _GXDataInfo()
                    parameters = _GXCommon.getData(data, info)
                if xml:
                    xml.appendStartTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR_WITH_SELECTION)
                    xml.appendStartTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR)
                    xml.appendComment(ci.__str__())
                    xml.appendLine(TranslatorTags.CLASS_ID, "Value", xml.integerToHex(ci.value, 4))
                    xml.appendComment(_GXCommon.toLogicalName(ln))
                    xml.appendLine(TranslatorTags.INSTANCE_ID, "Value", GXByteBuffer.hex(ln, False))
                    xml.appendLine(TranslatorTags.ATTRIBUTE_ID, "Value", xml.integerToHex(attributeIndex, 2))
                    xml.appendEndTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR)
                    xml.appendEndTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR_WITH_SELECTION)
                else:
                    obj = settings.objects.findByLN(ci, _GXCommon.toLogicalName(ln))
                    if obj is None:
                        obj = server.onFindObject(ci, 0, _GXCommon.toLogicalName(ln))
                    if obj is None:
                        e = ValueEventArgs(server, obj, attributeIndex, 0, 0)
                        e.error = ErrorCode.UNDEFINED_OBJECT
                        list_.append(e)
                    else:
                        arg = ValueEventArgs(server, obj, attributeIndex, selector, parameters)
                        arg.invokeId = (invokeID)
                        if server.onGetAttributeAccess(arg) == AccessMode.NO_ACCESS:
                            arg.error = ErrorCode.READ_WRITE_DENIED
                            list_.append(arg)
                        else:
                            list_.append(arg)
                pos += 1
            cnt = _GXCommon.getObjectCount(data)
            if xml:
                xml.appendEndTag(TranslatorTags.ATTRIBUTE_DESCRIPTOR_LIST)
                xml.appendStartTag(TranslatorTags.VALUE_LIST, "Qty", xml.integerToHex(cnt, 2))
            pos = 0
            while pos != cnt:
                di = _GXDataInfo()
                di.xml = xml
                if xml and xml.outputType == TranslatorOutputType.STANDARD_XML:
                    xml.appendStartTag(Command.WRITE_REQUEST, SingleReadResponse.DATA)
                value = _GXCommon.getData(data, di)
                if not di.complete:
                    value = GXByteBuffer.hex(data.data, False, data.position, len(data) - data.position)
                elif isinstance(value, bytearray):
                    value = GXByteBuffer.hex(value, False)
                if xml and xml.outputType == TranslatorOutputType.STANDARD_XML:
                    xml.appendEndTag(Command.WRITE_REQUEST, SingleReadResponse.DATA)
                pos += 1
            if xml:
                xml.appendEndTag(TranslatorTags.VALUE_LIST)
        except Exception as ex:
            if xml is None:
                raise ex

    @classmethod
    def handleMethodRequest(cls, settings, server, data, connectionInfo, replyData, xml):
        error = ErrorCode.OK
        bb = GXByteBuffer()
        type_ = data.getUInt8()
        invokeId = data.getUInt8()
        settings.updateInvokeId(invokeId)
        ci = data.getUInt16()
        ot = ObjectType(ci)
        ln = bytearray(6)
        data.get(ln)
        id_ = data.getUInt8()
        parameters = None
        selection = data.getUInt8()
        if xml:
            xml.appendStartTag(Command.METHOD_REQUEST)
            if type_ == ActionRequestType.NORMAL:
                xml.appendStartTag(Command.METHOD_REQUEST, ActionRequestType.NORMAL)
                xml.appendLine(TranslatorTags.INVOKE_ID, "Value", xml.integerToHex(invokeId, 2))
                cls.appendMethodDescriptor(xml, ci, ln, id_)
                if selection != 0:
                    xml.appendStartTag(TranslatorTags.METHOD_INVOCATION_PARAMETERS)
                    di = _GXDataInfo()
                    di.xml = (xml)
                    _GXCommon.getData(data, di)
                    xml.appendEndTag(TranslatorTags.METHOD_INVOCATION_PARAMETERS)
                xml.appendEndTag(Command.METHOD_REQUEST, ActionRequestType.NORMAL)
            xml.appendEndTag(Command.METHOD_REQUEST)
            return
        if selection != 0:
            info = _GXDataInfo()
            parameters = _GXCommon.getData(data, info)
        obj = settings.objects.findByLN(ot, _GXCommon.toLogicalName(ln))
        if not settings.acceptConnection() and (ci != ObjectType.ASSOCIATION_LOGICAL_NAME or id_ != 1):
            replyData.set(server.generateConfirmedServiceError(ConfirmedServiceError.INITIATE_ERROR, ServiceError.SERVICE, Service.UNSUPPORTED))
            return
        if obj is None:
            obj = server.onFindObject(ot, 0, _GXCommon.toLogicalName(ln))
        if obj is None:
            error = ErrorCode.UNDEFINED_OBJECT
        else:
            e = ValueEventArgs(server, obj, id_, 0, parameters)
            e.invokeId = (invokeId)
            if server.onGetMethodAccess(e) == MethodAccessMode.NO_ACCESS:
                error = ErrorCode.READ_WRITE_DENIED
            else:
                server.onPreAction(list(e))
                if e.handled:
                    actionReply = int(e.value)
                else:
                    actionReply = obj.invoke(settings, e)
                server.onPostAction(list(e))
                if actionReply and e.error == ErrorCode.OK:
                    bb.setUInt8(1)
                    bb.setUInt8(0)
                    if e.byteArray:
                        bb.set(actionReply)
                    else:
                        _GXCommon.setData(bb, _GXCommon.getDLMSDataType(actionReply), actionReply)
                else:
                    error = e.error
                    bb.setUInt8(0)
                invokeId = int(e.invokeId)
        p = GXDLMSLNParameters(settings, invokeId, Command.METHOD_RESPONSE, 1, None, bb, error)
        GXDLMS.getLNPdu(p, replyData)
        if isinstance(obj, (GXDLMSAssociationLogicalName,)) and id_ == 1:
            if (obj).getAssociationStatus() == AssociationStatus.ASSOCIATED:
                server.notifyConnected(connectionInfo)
                settings.setConnected(settings.connected | ConnectionState.DLMS)
            else:
                server.onInvalidConnection(connectionInfo)
                settings.setConnected(settings.connected & ~ConnectionState.DLMS)

    @classmethod
    def handleAccessRequest(cls, settings, server, data, reply, xml):
        #pylint: disable=bad-option-value,redefined-variable-type
        if xml is None and not settings.acceptConnection():
            reply.set(server.generateConfirmedServiceError(ConfirmedServiceError.INITIATE_ERROR, ServiceError.SERVICE, Service.UNSUPPORTED))
            return
        invokeId = data.getUInt32()
        settings.setLongInvokeID(invokeId)
        len_ = _GXCommon.getObjectCount(data)
        tmp = None
        if len_ != 0:
            tmp = bytearray(len_)
            data.get(tmp)
            if xml is None:
                dt = DataType.DATETIME
                if len_ == 4:
                    dt = DataType.TIME
                elif len_ == 5:
                    dt = DataType.DATE
                info = _GXDataInfo()
                info.type = dt
                _GXCommon.getData(GXByteBuffer(tmp), info)
        cnt = _GXCommon.getObjectCount(data)
        if xml:
            xml.appendStartTag(Command.ACCESS_REQUEST)
            xml.appendLine(TranslatorTags.LONG_INVOKE_ID, "Value", xml.integerToHex(invokeId, 8))
            xml.appendLine(TranslatorTags.DATE_TIME, "Value", GXByteBuffer.hex(tmp, False))
            xml.appendStartTag(TranslatorTags.ACCESS_REQUEST_BODY)
            xml.appendStartTag(TranslatorTags.LIST_OF_ACCESS_REQUEST_SPECIFICATION, "Qty", xml.integerToHex(cnt, 2))
        type_ = 0
        pos = 0
        while pos != cnt:
            type_ = AccessServiceCommandType(data.getUInt8())
            if type_ not in (AccessServiceCommandType.GET, AccessServiceCommandType.SET, AccessServiceCommandType.ACTION):
                raise ValueError("Invalid access service command type.")
            ci = data.getUInt16()
            ln = bytearray(6)
            data.get(ln)
            attributeIndex = data.getUInt8()
            if xml:
                xml.appendStartTag(TranslatorTags.ACCESS_REQUEST_SPECIFICATION)
                xml.appendStartTag(Command.ACCESS_REQUEST, type_)
                cls.appendAttributeDescriptor(xml, ci, ln, attributeIndex)
                xml.appendEndTag(Command.ACCESS_REQUEST, type_)
                xml.appendEndTag(TranslatorTags.ACCESS_REQUEST_SPECIFICATION)
            pos += 1
        if xml:
            xml.appendEndTag(TranslatorTags.LIST_OF_ACCESS_REQUEST_SPECIFICATION)
            xml.appendStartTag(TranslatorTags.ACCESS_REQUEST_LIST_OF_DATA, "Qty", xml.integerToHex(cnt, 2))
        cnt = _GXCommon.getObjectCount(data)
        pos = 0
        while pos != cnt:
            di = _GXDataInfo()
            di.xml = xml
            if xml and xml.outputType == TranslatorOutputType.STANDARD_XML:
                xml.appendStartTag(Command.WRITE_REQUEST, SingleReadResponse.DATA)
            value = _GXCommon.getData(data, di)
            if not di.complete:
                value = GXByteBuffer.hex(data.data, False, data.position, len(data) - data.position)
            elif isinstance(value, bytearray):
                value = GXByteBuffer.hex(value, False)
            if xml and xml.outputType == TranslatorOutputType.STANDARD_XML:
                xml.appendEndTag(Command.WRITE_REQUEST, SingleReadResponse.DATA)
            pos += 1
        if xml:
            xml.appendEndTag(TranslatorTags.ACCESS_REQUEST_LIST_OF_DATA)
            xml.appendEndTag(TranslatorTags.ACCESS_REQUEST_BODY)
            xml.appendEndTag(Command.ACCESS_REQUEST)

    @classmethod
    def handleEventNotification(cls, settings, reply, list_):
        data = reply.data
        reply.time = None
        len_ = data.getUInt8()
        tmp = None
        if len_ != 0:
            tmp = bytearray(len_)
            data.get(tmp)
            reply.time = _GXCommon.changeType(tmp, DataType.DATETIME)
        if reply.xml:
            reply.xml.appendStartTag(Command.EVENT_NOTIFICATION)
            if reply.time:
                reply.xml.appendComment(str(reply.time))
                reply.xml.appendLine(TranslatorTags.TIME, None, GXByteBuffer.hex(tmp, False))
        ci = data.getUInt16()
        ln = bytearray(6)
        data.get(ln)
        index = data.getUInt8()
        if reply.xml:
            cls.appendAttributeDescriptor(reply.xml, ci, ln, index)
            reply.xml.appendStartTag(TranslatorTags.ATTRIBUTE_VALUE)
        di = _GXDataInfo()
        di.xml = reply.xml
        value = _GXCommon.getData(reply.data, di)
        if reply.xml:
            reply.xml.appendEndTag(TranslatorTags.ATTRIBUTE_VALUE)
            reply.xml.appendEndTag(Command.EVENT_NOTIFICATION)
        else:
            obj = settings.objects.findByLN(ObjectType(ci), _GXCommon.toLogicalName(ln))
            if obj:
                v = ValueEventArgs(obj, index, 0, None)
                v.value = value
                obj.setValue(settings, v)
                list_.append((obj, int(index)))
            else:
                print("InformationReport message. Unknown object : " + str(ObjectType(ci)) + " " + _GXCommon.toLogicalName(ln))
